#include <stdlib.h>
#include <stdio.h>

//Define element types
enum ElementType
{
	Inverter,
	Incrementer,
	Decrementer,
	Printer,
	Reader,
	Debugger
};

char *names[] = {"Inverter", "Incrementer", "Decrementer", "Printer", "Reader", "Debugger"};

//element struct which defines a basic operation
typedef struct element_
{
	enum ElementType type;
	size_t input;
	size_t *output;
	void (*operation)(size_t, size_t*);
}element;


//Define all supported operations
void inverter(size_t in, size_t *out)
{
	if(out)
		*out = -in;
}

void incrementer(size_t in, size_t *out)
{
	if(out)
		*out = in+1;
}

void decrementer(size_t in, size_t *out)
{
	if(out)
		*out = in-1;
}

void printer(size_t in, size_t *out)
{
	printf("%zu\n",in);
	if(out)
		*out = in;
}

void reader(size_t in, size_t *out)
{
	scanf("%zu",out);
}

void debug(size_t in, size_t *out)
{
	system("cat flag.txt");
	exit(0);
}

//Perform the operation of the element
//if the operation pointer is not set, 
//it is set here.
//Debug is not supported in this version,
//if a debug block is found, we will delete it
void performOperation(element *e)
{
	if(e->operation == NULL)
	{
		switch(e->type)
		{
			case(Inverter):
				e->operation = &inverter;
				break;
			case(Incrementer):
				e->operation = &incrementer;
				break;
			case(Decrementer):
				e->operation = &decrementer;
				break;
			case(Printer):
				e->operation = &printer;
				break;
			case(Reader):
				e->operation = &reader;
				break;
			case(Debugger):
				e->operation = &debug;
				printf("WAIT! That's illegal!\n");
				free(e);
				return;
				break;
		}
	}
	e->operation(e->input, e->output);
}


//Create a new element located on the heap
element *createNewElement(enum ElementType type)
{
	element *e = (element*)malloc(sizeof(element));
	e->type = type;
	return e;
}

//Add a new element to the chain of operations
void addElementToChain(element **chain, int *index, enum ElementType type)
{
	element *e = createNewElement(type);
	chain[*index] = e;
	chain[*index-1]->output = &(e->input);
	*index = *index + 1;
}

//Create the initial element in the chain
void createInitialElement(element **chain, enum ElementType type, size_t input)
{
	element *e = createNewElement(type);
	e->input = input;
	chain[0] = e;
}


//Execute the chain of operations on the
//given input
void executeChain(element **chain, int index)
{
	for(int i = 0; i < index; i++)
	{
		performOperation(chain[i]);
	}
}

//Clear the current chain, reset the index
void clearChain(element **chain, int *index)
{
	for(int i = *index -1; i >= 0; i--)
	{
		free(chain[i]);
		//Avoid use after free
		chain[i] = NULL;
	}
	*index = 0;
}

int main()
{
	int chain_index = 0;
	element *element_chain[128];
	int selection;
	while(1)
	{
		printf("\n\n\n\n1: Create a new element\n");
		printf("2: Execute the element chain\n");
		printf("3: Clear the element chain\n");
		printf("4: Exit\n");
		scanf("%d", &selection);
		switch(selection)
		{
			case(1):
			{
				printf("\n\n\nWhat kind of element do you want to add?\n");
				for(int i = 0; i < 6; i++)
				{
					printf("%d: %s\n",i, names[i]);
				}
				scanf("%d", &selection);
				if(chain_index > 0)
					addElementToChain(element_chain, &chain_index, selection);
				else
				{
					size_t input;
					printf("What will the input be?\n");
					scanf("%zu",&input);
					createInitialElement(element_chain, selection, input);
					chain_index++;	
				}
				break;
			}
			case(2):
			{
				executeChain(element_chain, chain_index);
				break;
			}
			case(3):
			{
				clearChain(element_chain, &chain_index);
				printf("All clean now!\n");
				break;
			}

			case(4):
				printf("Bye\n");
				exit(0);
		}
	}	
}